/**
 * ROSC random number routines
 * void rosc_setup(void) -- increases the frequency of the ROSC
 * uint32_t rand_rosc_VN(void) -- returns 32-bit unsigned int number (using Von Neumann extractor)
 * uint32_t rand_rosc(void) -- returns 32-bit unsigned int number 
 * s15x16 uniform_rand(void) -- returns a fixed point number uniformly dist between -1 and +1
 * s15x16 normal_rand(void) -- returns a fixed point number normally dist with mean 0 and std-dev. 1.0
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 *  - GPIO 18 ---> 330 ohm resistor ---> VGA Green lo-bit |__ both wired to 150 ohm to ground 
 *  - GPIO 19 ---> 220 ohm resistor ---> VGA Green hi_bit |   and to VGA Green
 *  - GPIO 20 ---> 330 ohm resistor ---> VGA Blue
 *  - GPIO 21 ---> 330 ohm resistor ---> VGA Red
 *  - RP2040 GND ---> VGA GND
 *
 * RESOURCES USED
 *  - PIO state machines 0, 1, and 2 on PIO instance 0
 *  - DMA channels 0, 1, 2, and 3
 *  - 153.6 kBytes of RAM (for pixel color data)
 *
 * Protothreads v1.1.1
 * Threads:
 * core 0:
 * histogram of random numbers
 * blink LED25 
 * Serial i/o 
 * There are five random number routines
 * 
 * 32-bit integer
rand_rosc and rand_rosc_VN

fixed s15x16 uniform -1 to 1
s15x16 uniform_rand(void)

 fixed s15x16 normal mean=0 SD=1
s15x16 normal_rand(void){

single bit low quality
random_bit, random_sign

single bit medium
random_bit_VN

single bit best
random_bit_rand
 */
// ==========================================
// === VGA graphics library
// ==========================================
#include "vga16_graphics.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/dma.h"
// // Our assembled programs:
// // Each gets the name <pio_filename.pio.h>
// #include "hsync.pio.h"
// #include "vsync.pio.h"
// #include "rgb.pio.h"

// ==========================================
// === protothreads setup
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "hardware/clocks.h"
#include "hardware/structs/clocks.h"
#include "hardware/regs/rosc.h"
#include "string.h"
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"
// flag between serial and compute threads
int new_value ;
// number of rands to avareage to make gaussian
// and make one pixel worth more counts
int scale ;
char dist_type[5] ;

// ==========================================
// === fixed point s15x16
// ==========================================
// s15x16 fixed point macros ==
// == resolution 2^-16 = 1.5e-5
// == dynamic range is 32767/-32768
typedef signed int s15x16;
#define muls15x16(a,b) ((s15x16)(((( signed long long )(a))*(( signed long long )(b)))>>16)) //multiply two fixed 16:16
#define float_to_s15x16(a) ((s15x16)((a)*65536.0)) // 2^16
#define s15x16_to_float(a) ((float)(a)/65536.0)
#define s15x16_to_int(a)    ((int)((a)>>16))
#define int_to_s15x16(a)    ((s15x16)((a)<<16))
#define divs15x16(a,b) ((s15x16)((((signed long long)(a)<<16)/(b)))) 
#define abss15x16(a) abs(a)

// ==========================================
// === Ring Osc random bit RNG
// setup for higher speed oscillator
// ==========================================
void rosc_setup(void){
  volatile uint32_t *rosc_div = (uint32_t *)(ROSC_BASE + ROSC_DIV_OFFSET) ;
  volatile uint32_t *rosc_ctl = (uint32_t *)(ROSC_BASE + ROSC_CTRL_OFFSET) ;
  volatile uint32_t *rosc_freqA = (uint32_t *)(ROSC_BASE + ROSC_FREQA_OFFSET) ;
  volatile uint32_t *rosc_freqB = (uint32_t *)(ROSC_BASE + ROSC_FREQB_OFFSET) ;
  
  // set divider to one for frequency measurement
  *rosc_div = ROSC_DIV_VALUE_PASS + 1 ;
  // speed up the ROSC so more cycles between reads
  // (dont use ROSC_CTRL_FREQ_RANGE_VALUE_TOOHIGH)
  // Measured at 241 MHz with theses settings
  *rosc_ctl =  ROSC_CTRL_FREQ_RANGE_VALUE_HIGH ;// | ROSC_CTRL_ENABLE_VALUE_ENABLE;
  *rosc_freqA = (ROSC_FREQA_PASSWD_VALUE_PASS<<16) | 0xffff ;
  *rosc_freqB = (ROSC_FREQB_PASSWD_VALUE_PASS<<16) | 0xffff ;
}

// diagnostics
uint32_t rand_count, rand_count_VN, extractor_count;

// ==========================================================
// good quality 32-bit TRNG
// ==========================================================
// Excution time is 25-35 uSec, mostly becuause of the Von Neumann extactor
// This is going to be more useful for seed generation for rand()
// ==========================================================
// basic random number gen with Von Neumann extractor
// and a small delay in the extractor
// AND one pass through rand() function
// https://en.wikipedia.org/wiki/Randomness_extractor
uint32_t rand_rosc_VN(void){
    int k, random ;
    int random_bit1, random_bit2 ;
    volatile uint32_t *rnd_reg = (uint32_t *)(ROSC_BASE + ROSC_RANDOMBIT_OFFSET);
    volatile uint32_t *rosc_ctl = (uint32_t *)(ROSC_BASE + ROSC_CTRL_OFFSET) ;
    
    for(k=0;k<32;k++){
      // von Neumann bit extractor         
      while(1){
        //extractor_count++ ;
        random_bit1=0x00000001 & (*rnd_reg);
        //  a small delay decorrelates the bits
        asm("nop") ; asm("nop") ;
        asm("nop") ; asm("nop") ;
        asm("nop") ; asm("nop") ;
        asm("nop") ; asm("nop") ;      
        random_bit2=0x00000001 & (*rnd_reg);
        // if the two are diferent, use the first one
        if(random_bit1!=random_bit2) break;
      }  
      // build the 32 bit sample
      random=(random << 1) | random_bit1 ;
    }
    srand(random) ;
    //rand_count++ ;
    random = rand() ;
    return random;
}

// ==========================================================
//  medium quality 32-bit TRNG
// ==========================================================
// Excution time is 6 uSec
// Plenty good enough for game init with the fast RSOC clock
// ==========================================================
// and a small delay in the bit extractor
// AND one pass through rand() function
uint32_t rand_rosc(void){
    int k, random ;
    int random_bit1 ;
    volatile uint32_t *rnd_reg = (uint32_t *)(ROSC_BASE + ROSC_RANDOMBIT_OFFSET);
    
    for(k=0;k<32;k++){    
      // at least one nop need to be here for timing  
      //extractor_count++ ;
      asm("nop") ; asm("nop") ; // adds one microsec to execution
      //asm("nop") ; asm("nop") ;
      random_bit1=0x00000001 & (*rnd_reg);
      random=(random << 1) | random_bit1 ;
    }
    //rand_count++ ;
    srand(random) ;
    random = rand() ;
    return random;
}

// ==========================================================
//  fixed s15x16 uniform -1 to 1
// ==========================================================
// Excution time is 5 uSec
// ==========================================================
#define s15x16_one 0x00010000
// AND one pass through rand() function
s15x16 uniform_rand(void){
    int k, random ;
    int random_bit1 ;
    s15x16 uniform ;
    volatile uint32_t *rnd_reg = (uint32_t *)(ROSC_BASE + ROSC_RANDOMBIT_OFFSET);
    //
    for(k=0; k<32; k++){    
      // at least one nop need to be here for timing  
      asm("nop") ; asm("nop") ; // adds one microsec to execution
      //asm("nop") ; asm("nop") ;
      random_bit1=0x00000001 & (*rnd_reg) ;
      random=(random << 1) | random_bit1 ;
    }
    srand(random) ;
    random = rand() ;
    uniform = (s15x16)(random & 0x1ffff) - s15x16_one ;
    return uniform;
}

// ==========================================================
//  fixed s15x16 normal mean=0 SD=1
//  uses 12 uniform nums
// ==========================================================
// Excution time is --- uSec
// ==========================================================
#define s15x16_one 0x00010000
#define s15x16_six 0x00060000
// 
s15x16 normal_rand(void){
    int k, j, random ;
    int random_bit1 ;
    s15x16 uniform=0, gaussian ;
    volatile uint32_t *rnd_reg = (uint32_t *)(ROSC_BASE + ROSC_RANDOMBIT_OFFSET);
    //
    for(j=0; j<12; j++){  
        for(k=0; k<32; k++){    
          // at least one nop need to be here for timing  
          asm("nop") ; asm("nop") ; // adds one microsec to execution
          //asm("nop") ; asm("nop") ;
          random_bit1=0x00000001 & (*rnd_reg) ;
          random=(random << 1) | random_bit1 ;
        }
      srand(random) ;
      random = rand() ;
      uniform += (s15x16)(random & 0x0000ffff)  ;
    }
    gaussian = uniform - s15x16_six ;
    return gaussian;
}

// ================================================
// single bit routines
// ================================================
volatile uint32_t *rnd_reg = (uint32_t *)(ROSC_BASE + ROSC_RANDOMBIT_OFFSET);
#define random_bit (0x00000001 & (*rnd_reg)) 
//
#define random_sign (((0x00000001 & (*rnd_reg))<<1)-1)
//
inline uint32_t random_bit_VN(void) {
  uint32_t random_bit1=0, random_bit2=0 ; 
  while(random_bit1==random_bit2){ 
      random_bit1=0x00000001 & (*rnd_reg);
     // asm("nop") ; asm("nop") ; 
     // asm("nop") ; asm("nop") ; 
     // asm("nop") ; asm("nop") ; 
      asm("nop") ; asm("nop") ; 
      random_bit2=0x00000001 & (*rnd_reg); 
      //if(random_bit1!=random_bit2) break; 
  }  
  return random_bit1 ; 
} 

inline uint32_t random_bit_rand(void) {
  uint32_t random_bit1=0, random_bit2=0 ; 
  random_bit1 = rand_rosc() & 0x01 ;
  return random_bit1 ; 
} 

// ==================================================
// === graphics demo -- RUNNING on core 0
// ==================================================
short int histogram[512] ;
//
static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);
    static s15x16 rand_test ; 
    float rand_sum, rand_sum_sq  ;
    static int begin_time, end_time ;
    
    // the protothreads interval timer
    PT_INTERVAL_INIT() ;

    // Draw some filled rectangles
    fillRect(64, 0, 176, 50, BLUE); // blue box
    fillRect(250, 0, 176, 50, RED); // red box
    fillRect(435, 0, 176, 50, GREEN); // green box

    // Write some text
    setTextColor(WHITE) ;
    setCursor(65, 0) ;
    setTextSize(1) ;
    writeString("Raspberry Pi Pico") ;
    setCursor(65, 10) ;
    writeString("Hardware ROSC random demo") ;
    setCursor(65, 20) ;
    writeString("Cornell ECE4760") ;

    setCursor(445, 10) ;
    setTextColor(BLACK) ;
    setTextSize(1) ;
    writeString("Protothreads rp2040 v1.1.1") ;
    setCursor(445, 20) ;
    writeString("16 color VGA ") ;
    setTextColor(WHITE) ;
    
    while(true) {
        // new_value set by serial thread
        PT_YIELD_UNTIL(pt, new_value==true) ;
        // erase old data
        fillRect(0, 51, 639, 430, BLACK); // black box
        fillRect(250, 0, 176, 50, RED); // red box
        memset(histogram, 0, sizeof(histogram)) ;
        rand_count = 0 ;
        rand_count_VN = 0 ;
        rand_sum = 0 ;
        rand_sum_sq = 0 ;
        
        //
        begin_time = PT_GET_TIME_usec() ;
        //
        
        // ===========================================================
        // TEST UNIFORM FIX RANDOM
        if (dist_type[0] == 'u'){
          while((histogram[s15x16_to_int(rand_test)]>>scale) <= 400){
              rand_test = uniform_rand();
              //printf("rand=%f \n", s15x16_to_float(rand_test) ) ;
              rand_sum += rand_test ;
              rand_count++ ;
              // scale up for plotting
              rand_test = muls15x16(rand_test + s15x16_one, int_to_s15x16(256)) ;
              histogram[s15x16_to_int(rand_test)]++ ;       
              drawPixel(50 + s15x16_to_int(rand_test), 479-(histogram[s15x16_to_int(rand_test)]>>scale), MED_GREEN) ;
          }
        
          end_time = PT_GET_TIME_usec() - begin_time ;
          printf("rand count=%d  usec=%d  usec/rand=%d\n", rand_count, end_time, end_time/rand_count) ;
          printf("mean=%f \n", s15x16_to_float(rand_sum)/(float)(rand_count) ) ;
        }
        // END TEST UNIFORM
        // ===========================================================
        
    
        // ===========================================================
        // TEST NORMAL RANDOM
        if(dist_type[0] == 'n'){
          while((histogram[s15x16_to_int(rand_test)]>>scale) <= 400){
            //while(rand_count <= 40000){
              rand_test = normal_rand();
              //printf("rand=%f \n", s15x16_to_float(rand_test) ) ;
              rand_sum += s15x16_to_float(rand_test) ;
              rand_sum_sq += s15x16_to_float(rand_test) * s15x16_to_float(rand_test) ;
              rand_count++ ;
              // scale up for plotting
              // divide by 6 because adding 12 uniforms to get gaussian
              rand_test = muls15x16(rand_test + s15x16_six, float_to_s15x16(256/6)) ;
              histogram[s15x16_to_int(rand_test)]++ ;       
              drawPixel(50 + s15x16_to_int(rand_test), 479-(histogram[s15x16_to_int(rand_test)]>>scale), MED_GREEN) ;
          }
        
          end_time = PT_GET_TIME_usec() - begin_time ;
          printf("rand count=%d  usec=%d  usec/rand=%d\n", rand_count, end_time, end_time/rand_count) ;
          float mean = (rand_sum)/(float)(rand_count) ;
          float var =  (rand_sum_sq) - (rand_sum * rand_sum)/ (float)(rand_count)  ;
          var = var/(float)(rand_count-1.0 ) ;
          float std_dev = sqrt(var);
          printf("mean=%f std.dev.=%f\n", (mean) , std_dev) ;

          // fit gaussian
          // (1/(sigma*sqrt(2*pi))) exp[-0.5((x-mean)/sigma)^2] scaled to number of events
          // mean = 512*n_gauss/2   and sigma = sqrt(512*n_gauss/12)
          
          float x, y ;
          float sqrt2pi = sqrt(2*3.14159) ;
          
          for(int i =0; i<1000; i++){
              x = (float) i/100 - 5 ;        
              y = 0.0235 *(float)rand_count * exp(-0.5*(x-mean)*(x-mean)/(var)) / (std_dev*sqrt2pi) ;
              //drawPixel(50 + (short)x, 479-(short)(y), RED);
              // fillCircle(short x0, short y0, short r, char color) ;
              fillCircle(50 + (short)(x*42.67+252), 479-((short)(y)>>scale), 1, RED);
          } 
        }
        // END TEST NORMAL
        // ===========================================================
      

       // =============================================================
       // Test binomial dist of single bits
        int n_binomial = 20 ;
        char vid_string[30];
        // binomial probabilities for n=20, fair coin
        float binomial[21] = {0.0000,0.0000,0.0002,0.0011,0.0046,0.0148,0.0370,0.0739,0.1201,0.1602,
                              0.1762,0.1602,0.1201,0.0739,0.0370,0.0148,0.0046,0.0011,0.0002,0.0000,0.0000} ;
        setTextColor2(CYAN, BLACK) ;
        if(dist_type[0] == 'b'){
          
          memset(histogram, 0, sizeof(histogram)) ;
          rand_count = 0; 
          rand_test = 0;
          
          while((histogram[rand_test]>>scale) <= 400){
              //
              rand_test = 0 ;
              for(int i=1; i<=n_binomial; i++){
                if(dist_type[1]=='l') rand_test += random_bit ;
                if(dist_type[1]=='m') rand_test += random_bit_VN() ;
                if(dist_type[1]=='h') rand_test += random_bit_rand() ;
              }
              
              //
              rand_count++ ;
              histogram[rand_test]++ ;       
              //drawPixel(50 + 10*rand_test, 479-(histogram[rand_test]>>scale), MED_GREEN);
              // fillRect(short x, short y, short w, short h, char color) ;
              fillRect(50 + 20*rand_test, 479-(histogram[rand_test]>>scale), 10, (histogram[rand_test]>>scale), GREEN) ;
          }  
          if(dist_type[1]=='l' | dist_type[1]=='m' | dist_type[1]=='h'){
            end_time = PT_GET_TIME_usec() - begin_time ;
            printf("rand count=%d  usec=%d  usec/sample(incl graphics)=%d\n", rand_count, end_time, end_time/rand_count) ;

            for(int i=0; i<=n_binomial; i++){
                sprintf(vid_string, "%02d", i);
                setCursor( 48 + 20*i, 455) ;
                writeString(vid_string) ;
                // plot expected
                fillCircle(55 + 20*i, 479-(short)(rand_count*binomial[i]*pow(2,-scale)), 3, RED) ;
            }
          }
        }
       // END TEST BinoMiAL
        // ===========================================================

        // clear execute flag and wait for new serial input
        new_value = false ;
            
   }
   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ==================================================
// === user's serial input thread on core 0
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      //
      while(1) {
        // print prompt
        sprintf(pt_serial_out_buffer, " number_of_counts/pixel, distrubution(u,n,bl,bm,bh): ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
         sscanf(pt_serial_in_buffer, "%d %s", &scale, dist_type) ;
         
         // start comp thread
         new_value = true ;

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");

  // speed up the rosc
    rosc_setup();
    // First pass TRNG sanity check
    printf("first pass TRNG %08x\n", rand_rosc()) ;
   // printf("test rand bits %2d %2d %2d %2d %2d\n", random_bit, random_bit, random_bit, random_bit, random_bit);
   // printf("test sign bits %2d %2d %2d %2d %2d\n", random_sign, random_sign, random_sign, random_sign, random_sign);
    // get actual ROSC speed
    printf("ROSC freq=%d MHz\n", frequency_count_khz(CLOCKS_FC0_SRC_VALUE_ROSC_CLKSRC)/1000) ;

    uint32_t head_count_VN=0, head_count=0;
    //
    for(int i=0; i<100; i++){
      head_count_VN += random_bit_VN() ;
      head_count += random_bit ;
    }
    printf("head count: %d %d\n", head_count, head_count_VN);

  // Initialize the VGA screen
  initVGA() ;
     
  // start core 1 threads
  //multicore_reset_core1();
  //multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_serial) ;
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main